"""Config models for shading rules."""
from __future__ import annotations

import copy
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class ShadingPosition(pydantic.BaseModel):
	"""Position of shading object"""
	position: float | bool | None = pydantic.Field(..., description="target position")
	slat: float | None = pydantic.Field(None, description="target slat position")

	def __init__(self, position: float | bool | None, slat: float | None = None) -> None:
		"""Initialize shading position with position and slat

		:param position: target position value
		:param slat: slat value
		"""
		super().__init__(position=position, slat=slat)


class ShadingItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for shading rules."""
	shading_position: HABApp.openhab.items.RollershutterItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="item for setting the shading position")
	slat: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="item for setting the slat value")
	manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to switch to manual mode and disable the automatic functions")
	shading_position_control: list[HABApp.openhab.items.RollershutterItem | HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="control items to improve manual detection")
	shading_position_group: list[HABApp.openhab.items.RollershutterItem | HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="")
	wind_alarm: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON when wind alarm is active")
	sun_protection: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON when sun protection is needed")
	sun_protection_slat: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="value for the slat when sun protection is active")
	sleeping_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="item of the sleeping state set via habapp_rules.system.sleep.Sleep")
	night: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON at night or darkness")
	door: HABApp.openhab.items.ContactItem | None = pydantic.Field(None, description="item for setting position when door is opened")
	summer: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item which is ON during summer")
	hand_manual_is_active_feedback: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="feedback item which is ON when hand or manual is active")
	state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the current state of the state machine")


class ShadingParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for shading rules."""
	pos_auto_open: ShadingPosition = pydantic.Field(ShadingPosition(0, 0), description="position for auto open")
	pos_wind_alarm: ShadingPosition | None = pydantic.Field(ShadingPosition(0, 0), description="position for wind alarm")
	pos_sleeping_night: ShadingPosition | None = pydantic.Field(ShadingPosition(100, 100), description="position for sleeping at night")
	pos_sleeping_day: ShadingPosition | None = pydantic.Field(None, description="position for sleeping at day")
	pos_sun_protection: ShadingPosition | None = pydantic.Field(ShadingPosition(100, None), description="position for sun protection")
	pos_night_close_summer: ShadingPosition | None = pydantic.Field(None, description="position for night close during summer")
	pos_night_close_winter: ShadingPosition | None = pydantic.Field(ShadingPosition(100, 100), description="position for night close during winter")
	pos_door_open: ShadingPosition | None = pydantic.Field(ShadingPosition(0, 0), description="position if door is opened")
	manual_timeout: int = pydantic.Field(24 * 3600, description="fallback timeout for manual state", gt=0)
	door_post_time: int = pydantic.Field(5 * 60, description="extended time after door is closed", gt=0)
	value_tolerance: int = pydantic.Field(0, description="value tolerance for shading position which is allowed without manual detection", ge=0)

	@pydantic.model_validator(mode="after")
	def validate_model(self) -> typing.Self:
		"""Validate model

		:return: validated model
		"""
		if self.pos_sleeping_night and not self.pos_sleeping_day:
			self.pos_sleeping_day = copy.deepcopy(self.pos_sleeping_night)
		return self


class ShadingConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for shading objects."""
	items: ShadingItems = pydantic.Field(..., description="items for shading")
	parameter: ShadingParameter = pydantic.Field(ShadingParameter(), description="parameter for shading")

	@pydantic.model_validator(mode="after")
	def validate_model(self) -> typing.Self:
		"""Validate model

		:return: validated model
		:raises AssertionError: if 'parameter.pos_night_close_summer' is set but 'items.summer' is missing
		"""
		if self.parameter.pos_night_close_summer is not None and self.items.summer is None:
			raise AssertionError("Night close position is set for summer, but item for summer / winter is missing!")
		return self


class ResetAllManualHandItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for reset all manual hand items."""
	reset_manual_hand: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item for resetting manual and hand state to automatic state")


class ResetAllManualHandParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for reset all manual hand parameter."""
	shading_objects: list[object] | None = pydantic.Field(None, description="list of shading objects to reset")


class ResetAllManualHandConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for reset all manual hand config."""
	items: ResetAllManualHandItems = pydantic.Field(..., description="items for reset all manual hand config")
	parameter: ResetAllManualHandParameter = pydantic.Field(ResetAllManualHandParameter(), description="parameter for reset all manual hand config")


class SlatValueItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for slat values for sun protection."""
	sun_elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item for sun elevation")
	slat_value: HABApp.openhab.items.NumberItem | HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="item for slat value, which should be set")
	summer: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item for summer")


class ElevationSlatMapping(pydantic.BaseModel):
	"""Mapping from elevation to slat value."""
	elevation: int
	slat_value: int

	def __init__(self, elevation: int, slat_value: int):
		"""Initialize the elevation slat mapping.

		:param elevation: elevation value
		:param slat_value: mapped slat value
		"""
		super().__init__(elevation=elevation, slat_value=slat_value)


class SlatValueParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for slat values for sun protection."""
	elevation_slat_characteristic: list[ElevationSlatMapping] = pydantic.Field([
		ElevationSlatMapping(0, 100),
		ElevationSlatMapping(4, 100),
		ElevationSlatMapping(8, 90),
		ElevationSlatMapping(18, 80),
		ElevationSlatMapping(26, 70),
		ElevationSlatMapping(34, 60),
		ElevationSlatMapping(41, 50)], description="list of tuple-mappings from elevation to slat value")
	elevation_slat_characteristic_summer: list[ElevationSlatMapping] = pydantic.Field([
		ElevationSlatMapping(0, 100),
		ElevationSlatMapping(4, 100),
		ElevationSlatMapping(8, 90),
		ElevationSlatMapping(18, 80)], description="list of tuple-mappings from elevation to slat value, which is used if summer is active")

	@pydantic.field_validator("elevation_slat_characteristic", "elevation_slat_characteristic_summer")
	@classmethod
	def sort_mapping(cls, values: list[ElevationSlatMapping]) -> list[ElevationSlatMapping]:
		"""Sort the elevation slat mappings.

		:param values: input values
		:return: sorted values
		:raises AssertionError: if elevation values are not unique
		"""
		values.sort(key=lambda x: x.elevation)

		if len(values) != len(set(value.elevation for value in values)):
			raise AssertionError("Elevation values must be unique!")

		return values


class SlatValueConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for slat values for sun protection."""
	items: SlatValueItems = pydantic.Field(..., description="items for slat values for sun protection")
	parameter: SlatValueParameter = pydantic.Field(SlatValueParameter(), description="parameter for slat values for sun protection")



"""Rule for evaluating a humidity sensor."""
import logging

import HABApp

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.humidity

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class HumiditySwitch(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Rule for setting humidity switch if high humidity or a high humidity change is detected."""

	states = [
		{"name": "off"},
		{"name": "on", "initial": "HighHumidity", "children": [
			{"name": "HighHumidity"},
			{"name": "Extended", "timeout": 99, "on_timeout": "on_extended_timeout"},
		]}
	]

	trans = [
		{"trigger": "high_humidity_start", "source": "off", "dest": "on"},
		{"trigger": "high_humidity_start", "source": "on_Extended", "dest": "on_HighHumidity"},
		{"trigger": "high_humidity_end", "source": "on_HighHumidity", "dest": "on_Extended"},
		{"trigger": "on_extended_timeout", "source": "on_Extended", "dest": "off"},
	]

	def __init__(self, config: habapp_rules.sensors.config.humidity.HumiditySwitchConfig) -> None:
		"""Init humidity rule.

		:param config: config for humidity switch rule
		"""
		self._config = config
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.humidity.name)

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self.state_machine.get_state("on_Extended").timeout = self._config.parameter.extended_time

		# register callbacks
		self._config.items.humidity.listen_event(self._cb_humidity, HABApp.openhab.events.ItemStateUpdatedEventFilter())

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
		super()._update_openhab_state()
		self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

		self._set_output()
		self._previous_state = self.state

	def _set_output(self) -> None:
		"""Set output."""
		target_state = "ON" if self.state in {"on_HighHumidity", "on_Extended"} else "OFF"
		habapp_rules.core.helper.send_if_different(self._config.items.output, target_state)

	def _get_initial_state(self, default_value: str = "initial") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		return "on" if self._check_high_humidity() else "off"

	def _check_high_humidity(self, humidity_value: float | None = None) -> bool:
		"""Check if humidity is above threshold.
		
		:param humidity_value: humidity value, which should be checked. If None, the value of the humidity item will be used
		:return: True if humidity is above threshold
		"""
		if humidity_value is None:
			if self._config.items.humidity.value is None:
				return False
			humidity_value = self._config.items.humidity.value

		return humidity_value >= self._config.parameter.absolute_threshold

	def _cb_humidity(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
		"""Callback, which is triggered if the humidity was updated.

		:param event: trigger event
		"""
		if self._check_high_humidity(event.value):
			self.high_humidity_start()
		else:
			self.high_humidity_end()



"""Base classes for pydantic config models."""
import types
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.exceptions
import habapp_rules.core.helper


class ItemBase(pydantic.BaseModel):
	"""Base class for item config."""
	model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

	@pydantic.model_validator(mode="before")
	@classmethod
	def check_all_fields_oh_items(cls, data: typing.Any) -> typing.Any:
		"""Validate that all fields are OpenHAB items.

		All items must be subclasses of `HABApp.openhab.items.OpenhabItem`.
		If create_if_not_exists is set, only one type is allowed.
		For lists, only one type is allowed.

		:param data: data object given by pydantic
		:return: data object
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if validation fails
		"""
		for name, field_info in cls.model_fields.items():
			field_types = cls._get_type_of_field(name)
			extra_args = extra if (extra := field_info.json_schema_extra) else {}

			if isinstance(field_types, types.GenericAlias):
				# type is list of OpenHAB items
				field_types = typing.get_args(field_types)[0]

				if isinstance(field_types, types.UnionType):
					field_types = [arg for arg in typing.get_args(field_types) if arg is not types.NoneType]

				# validate that create_if_not_exists is not set for lists
				if extra_args.get("create_if_not_exists", False):
					raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("create_if_not_exists is not allowed for list fields")

			field_types = field_types if isinstance(field_types, list) else [field_types]

			for field_type in field_types:
				if not issubclass(field_type, HABApp.openhab.items.OpenhabItem):
					raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"Field {field_type} is not an OpenhabItem")

			# validate that only one type is given if create_if_not_exists is set
			if extra_args.get("create_if_not_exists", False) and len(field_types) > 1:
				raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("If create_if_not_exists is set, only one type is allowed")

		return data

	@pydantic.field_validator("*", mode="before")
	@classmethod
	def convert_to_oh_item(cls, var: str | HABApp.openhab.items.OpenhabItem, validation_info: pydantic.ValidationInfo) -> HABApp.openhab.items.OpenhabItem | None:
		"""Convert to OpenHAB item.

		:param var: variable given by the user
		:param validation_info: validation info given by pydantic
		:return: variable converted to OpenHAB item
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if type is not supported
		"""

		extra_args = extra if (extra := cls.model_fields[validation_info.field_name].json_schema_extra) else {}
		create_if_not_exists = extra_args.get("create_if_not_exists", False)

		if create_if_not_exists:
			item_type = cls._get_type_of_field(validation_info.field_name).__qualname__.removesuffix("Item")
			return habapp_rules.core.helper.create_additional_item(var, item_type)

		if isinstance(var, list):
			return [cls._get_oh_item(itm) for itm in var]

		if issubclass(type(var), HABApp.openhab.items.OpenhabItem) or isinstance(var, str):
			return cls._get_oh_item(var)

		if var is None:
			return None

		raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"The following var is not supported: {var}")

	@staticmethod
	def _get_oh_item(item: str | HABApp.openhab.items.OpenhabItem) -> HABApp.openhab.items.OpenhabItem:
		"""get OpenHAB item from string or OpenHAB item

		:param item: name of OpenHAB item or OpenHAB item
		:return: OpenHAB item
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if type is not supported
		"""
		return item if isinstance(item, HABApp.openhab.items.OpenhabItem) else HABApp.openhab.items.OpenhabItem.get_item(item)

	@classmethod
	def _get_type_of_field(cls, field_name: str) -> type | list[type]:
		"""Get type of field

		:param field_name: name of field
		:return: type of field, NoneType will be ignored
		"""
		field_type = cls.model_fields[field_name].annotation
		if isinstance(field_type, types.UnionType):
			field_type = [arg for arg in typing.get_args(field_type) if arg is not types.NoneType]
		return field_type


class ParameterBase(pydantic.BaseModel):
	"""Base class for parameter config."""
	model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)


class ConfigBase(pydantic.BaseModel):
	"""Base class for config objects."""
	items: ItemBase | None
	parameter: ParameterBase | None

	def __init__(self, **data: typing.Any) -> None:
		"""Initialize the model.

		:param data: data object given by pydantic
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if validation fails
		"""
		try:
			super().__init__(**data)
		except pydantic.ValidationError as exc:
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException(f"Failed to validate model: {exc.errors()}")



"""Base class for Rule with State Machine."""
import threading

import HABApp
import HABApp.openhab.connection.handler.func_sync
import transitions.extensions.states



@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class StateMachineWithTimeout(transitions.Machine):
	"""State machine class with timeout"""


@transitions.extensions.states.add_state_features(transitions.extensions.states.Timeout)
class HierarchicalStateMachineWithTimeout(transitions.extensions.HierarchicalMachine):
	"""Hierarchical state machine class with timeout"""


class StateMachineRule(HABApp.Rule):
	"""Base class for creating rules with a state machine."""
	states: list[dict] = []
	trans: list[dict] = []
	state: str

	def __init__(self, state_item: HABApp.openhab.items.StringItem) -> None:
		"""Init rule with state machine.

		:param state_item: name of the item to hold the state
		"""
		self.state_machine: transitions.Machine | None = None
		HABApp.Rule.__init__(self)

		self._item_state = state_item

	def get_initial_log_message(self) -> str:
		"""Get log message which can be logged at the init of a rule with a state machine.

		:return: log message
		"""
		return f"Init of rule '{self.__class__.__name__}' with name '{self.rule_name}' was successful. Initial state = '{self.state}' | State item = '{self._item_state.name}'"

	def _get_initial_state(self, default_value: str = "initial") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_state.value and self._item_state.value in [item.get("name", None) for item in self.states if isinstance(item, dict)]:
			return self._item_state.value
		return default_value

	def _set_initial_state(self) -> None:
		"""Set initial state.
		if the ``initial_state`` parameter of the state machine constructor is used the timeouts will not be started for the initial state.
		"""
		self._set_state(self._get_initial_state())

	def _set_state(self, state_name: str) -> None:
		"""Set given state

		:param state_name: name of state
		"""
		eval(f"self.to_{state_name}()")  # pylint: disable=eval-used

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
		self._item_state.oh_send_command(self.state)

	def on_rule_removed(self) -> None:
		"""Override this to implement logic that will be called when the rule has been unloaded."""
		# stop timeout timer of current state
		if self.state_machine:
			for itm in self.state_machine.get_state(self.state).runner.values():
				if isinstance(itm, threading.Timer) and itm.is_alive():
					itm.cancel()



"""Config models for humidity rules."""
import HABApp
import pydantic

import habapp_rules.core.pydantic_base


class HumiditySwitchItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for humidity switch."""
	humidity: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item which holds the measured humidity")
	output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which will be switched on if high humidity is detected")
	state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the state")


class HumiditySwitchParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for humidity switch."""
	absolute_threshold: float = pydantic.Field(65, description="threshold for high humidity")
	extended_time: int = pydantic.Field(10 * 60, description="extended time in seconds, if humidity is below threshold")


class HumiditySwitchConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for humidity switch."""
	items: HumiditySwitchItems = pydantic.Field(..., description="items for humidity switch")
	parameter: HumiditySwitchParameter = pydantic.Field(HumiditySwitchParameter(), description="parameter for humidity switch")



"""Light HCL rules."""
import abc
import datetime
import logging

import HABApp

import habapp_rules.actors.config.light_hcl
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.core.type_of_day
import habapp_rules.system

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class _HclBase(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Base class for HCL rules."""

	states = [
		{"name": "Manual"},
		{"name": "Hand", "timeout": 99, "on_timeout": "hand_timeout"},
		{"name": "Auto", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "HCL"},
			{"name": "Sleep", "initial": "Active", "children": [
				{"name": "Active"},
				{"name": "Post", "timeout": 1, "on_timeout": "post_sleep_timeout"}
			]},
			{"name": "Focus"}
		]}
	]

	trans = [
		{"trigger": "manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
		{"trigger": "manual_off", "source": "Manual", "dest": "Auto"},

		{"trigger": "hand_on", "source": "Auto", "dest": "Hand"},
		{"trigger": "hand_timeout", "source": "Hand", "dest": "Auto"},

		{"trigger": "sleep_start", "source": ["Auto_HCL", "Auto_Focus"], "dest": "Auto_Sleep"},
		{"trigger": "sleep_end", "source": "Auto_Sleep_Active", "dest": "Auto_Sleep_Post"},
		{"trigger": "post_sleep_timeout", "source": "Auto_Sleep_Post", "dest": "Auto_HCL"},

		{"trigger": "focus_start", "source": ["Auto_HCL", "Auto_Sleep"], "dest": "Auto_Focus"},
		{"trigger": "focus_end", "source": "Auto_Focus", "dest": "Auto_HCL"},
	]

	def __init__(self, config: habapp_rules.actors.config.light_hcl.HclElevationConfig | habapp_rules.actors.config.light_hcl.HclTimeConfig) -> None:
		"""Init base class.

		:param config: config of HCL light.
		"""
		self._config = config

		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.color.name)

		self._state_observer = habapp_rules.actors.state_observer.StateObserverNumber(self._config.items.color.name, self._cb_hand, value_tolerance=10)

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self._set_timeouts()

		# set callbacks
		self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.sleep_state is not None:
			self._config.items.sleep_state.listen_event(self._cb_sleep, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.focus is not None:
			self._config.items.focus.listen_event(self._cb_focus, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.switch_on is not None:
			self._config.items.switch_on.listen_event(self._cb_switch_on, HABApp.openhab.events.ItemStateChangedEventFilter())

	def _set_timeouts(self) -> None:
		"""Set timeouts."""
		self.state_machine.get_state("Auto_Sleep_Post").timeout = self._config.parameter.post_sleep_timeout
		self.state_machine.get_state("Hand").timeout = self._config.parameter.hand_timeout

	def _get_initial_state(self, default_value: str = "") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._config.items.manual.is_on():
			return "Manual"
		if self._config.items.sleep_state is not None and self._config.items.sleep_state.value in (habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value):
			return "Auto_Sleep"
		if self._config.items.focus is not None and self._config.items.focus.is_on():
			return "Auto_Focus"
		return "Auto_HCL"

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item and other states.

		This should method should be set to "after_state_change" of the state machine.
		"""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

			self._set_light_color()
			self._previous_state = self.state

	def _set_light_color(self):
		"""Set light color."""
		target_color = None

		if self.state == "Auto_HCL":
			target_color = self._get_hcl_color()
		elif self.state == "Auto_Focus":
			target_color = self._config.parameter.focus_color
		elif self.state == "Auto_Sleep_Active":
			target_color = self._config.parameter.sleep_color

		if target_color is not None:
			self._state_observer.send_command(target_color)

	def on_enter_Auto_Init(self) -> None:  # pylint: disable=invalid-name
		"""Is called on entering of init state"""
		self._set_initial_state()

	@abc.abstractmethod
	def _get_hcl_color(self) -> int | None:
		"""Get HCL color.

		:return: HCL light color
		"""

	@staticmethod
	def _get_interpolated_value(config_start: tuple[float, float], config_end: tuple[float, float], value: float) -> float:
		"""Get interpolated value

		:param config_start: start config
		:param config_end: end config
		:param value: input value which is the input for the interpolation
		:return: interpolated value
		"""
		fit_m = (config_end[1] - config_start[1]) / (config_end[0] - config_start[0])
		fit_t = config_end[1] - fit_m * config_end[0]

		return fit_m * value + fit_t

	def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the manual switch has a state change event.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.manual_on()
		else:
			self.manual_off()

	def _cb_hand(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual change was detected.

		:param event: original trigger event
		"""
		self._instance_logger.debug("Hand detected")
		self.hand_on()

	def _cb_focus(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the focus switch has a state change event.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.focus_start()
		else:
			self.focus_end()

	def _cb_sleep(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the sleep state has a state change event.

		:param event: trigger event
		"""
		if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
			self.sleep_start()
			if self._config.items.focus is not None and self._config.items.focus.is_on():
				self._config.items.focus.oh_send_command("OFF")
		elif event.value == habapp_rules.system.SleepState.AWAKE.value:
			self.sleep_end()

	def _cb_switch_on(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the switch_on_item has a state change event.

		:param event: trigger event
		"""
		if self.state == "Auto_HCL" and event.value == "ON" or isinstance(event.value, (int, float)) and event.value > 0:
			if (target_color := self._get_hcl_color()) is not None:
				self.run.at(1, self._state_observer.send_command, target_color)


class HclElevation(_HclBase):
	"""Sun elevation based HCL.

	# Items:
	Number    Elevation                     "Elevation [%s]"                {channel="astro:sun:home:position#elevation"}
	Number    HCL_Color_Elevation           "HCL Color Elevation"
	Switch    HCL_Color_Elevation_manual    "HCL Color Elevation manual"

	# Config
	config = habapp_rules.actors.config.light_hcl.HclElevationConfig(
		items=habapp_rules.actors.config.light_hcl.HclElevationItems(
			elevation="Elevation",
			color="HCL_Color_Elevation",
			manual="HCL_Color_Elevation_manual",
		),
		parameter=habapp_rules.actors.config.light_hcl.HclElevationParameter(
			color_map=[(0, 2000), (10, 4000), (30, 5000)]
		)
	)

	# Rule init:
	habapp_rules.actors.light_hcl.HclElevation(config)
	"""

	def __init__(self, config: habapp_rules.actors.config.light_hcl.HclElevationConfig) -> None:
		"""Init sun elevation based HCL rule.

		:param config: config for HCL rule
		"""
		_HclBase.__init__(self, config)
		self._config = config

		self._config.items.elevation.listen_event(self._cb_elevation, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._cb_elevation(None)

	def _get_hcl_color(self) -> int | None:
		"""Get HCL color depending on elevation

		:return: HCL light color
		"""
		elevation = self._config.items.elevation.value

		if elevation is None:
			return None

		return_value = 0
		if elevation <= self._config.parameter.color_map[0][0]:
			return_value = self._config.parameter.color_map[0][1]

		elif elevation >= self._config.parameter.color_map[-1][0]:
			return_value = self._config.parameter.color_map[-1][1]

		else:
			for idx, config_itm in enumerate(self._config.parameter.color_map):  # pragma: no cover
				if config_itm[0] <= elevation <= self._config.parameter.color_map[idx + 1][0]:
					return_value = self._get_interpolated_value(config_itm, self._config.parameter.color_map[idx + 1], elevation)
					break

		return round(return_value)

	def _cb_elevation(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
		"""Callback which is called if elevation changed"""
		if self.state == "Auto_HCL" and self._config.items.elevation.value is not None:
			self._state_observer.send_command(self._get_hcl_color())


class HclTime(_HclBase):
	"""Time based HCL.
	# Items:
	Number    HCL_Color_Time           "HCL Color Time"
	Switch    HCL_Color_Time_manual    "HCL Color Time manual"

	# Config
	config = habapp_rules.actors.config.light_hcl.HclTimeConfig(
		items=habapp_rules.actors.config.light_hcl.HclTimeItems(
			color="HCL_Color_Time",
			manual="HCL_Color_Time_manual",
		),
		parameter=habapp_rules.actors.config.light_hcl.HclTimeParameter(
			[(6, 2000), (12, 4000), (20, 3000)],
		)
	)

	# Rule init:
	habapp_rules.actors.light_hcl.HclTime(config)
	"""

	def __init__(self, config: habapp_rules.actors.config.light_hcl.HclTimeConfig) -> None:
		"""Init time based HCL rule.

		:param config: config for HCL light rule
		"""
		_HclBase.__init__(self, config)
		self.run.every(None, 300, self._update_color)  # every 5 minutes

	def _one_hour_later(self, current_time: datetime.datetime) -> bool:
		"""Check if today the color values will be shifted one hour later in the evening

		:param current_time: current time
		:return: True if next day is a weekend / holiday day
		"""
		if not self._config.parameter.shift_weekend_holiday:
			return False

		if current_time.hour > 12 and (habapp_rules.core.type_of_day.is_holiday(1) or habapp_rules.core.type_of_day.is_weekend(1)):
			return True
		if current_time.hour <= 4 and (habapp_rules.core.type_of_day.is_holiday() or habapp_rules.core.type_of_day.is_weekend()):
			return True
		return False

	def _get_hcl_color(self) -> int | None:
		"""Get HCL color depending on time

		:return: HCL light color
		"""
		current_time = datetime.datetime.now()

		if self._one_hour_later(current_time):
			current_time -= datetime.timedelta(hours=1)

		if current_time.hour < self._config.parameter.color_map[0][0]:
			start_config = (self._config.parameter.color_map[-1][0] - 24, self._config.parameter.color_map[-1][1])
			end_config = self._config.parameter.color_map[0]

		elif current_time.hour >= self._config.parameter.color_map[-1][0]:
			start_config = self._config.parameter.color_map[-1]
			end_config = (self._config.parameter.color_map[0][0] + 24, self._config.parameter.color_map[0][1])

		else:
			for idx, config_itm in enumerate(self._config.parameter.color_map):  # pragma: no cover
				if config_itm[0] <= current_time.hour < self._config.parameter.color_map[idx + 1][0]:
					start_config = config_itm
					end_config = self._config.parameter.color_map[idx + 1]
					break

		return round(self._get_interpolated_value(start_config, end_config, current_time.hour + current_time.minute / 60))

	def _update_color(self) -> None:
		"""Callback which is called every 5 minutes"""
		if self.state == "Auto_HCL":
			self._state_observer.send_command(self._get_hcl_color())



"""Config models for monthly energy report."""
import HABApp
import multi_notifier.connectors.connector_mail
import pydantic

import habapp_rules.core.pydantic_base


class EnergyShare(pydantic.BaseModel):
	"""Dataclass for defining energy share objects."""
	model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

	energy_item: HABApp.openhab.items.NumberItem
	chart_name: str
	monthly_power: float = 0.0

	def __init__(self, energy_item: str | HABApp.openhab.items.NumberItem, chart_name: str, monthly_power: float = 0.0):
		"""Init energy share object without keywords.

		:param energy_item: name or item of energy
		:param chart_name: name which will be shown in the chart
		:param monthly_power: monthly power of this energy share. This will be set by the energy share rule.
		"""
		super().__init__(energy_item=energy_item, chart_name=chart_name, monthly_power=monthly_power)

	@pydantic.field_validator("energy_item", mode="before")
	@classmethod
	def check_oh_item(cls, data: str | HABApp.openhab.items.NumberItem) -> HABApp.openhab.items.NumberItem:
		"""Check if given item is an OpenHAB item or try to get it from OpenHAB.

		:param data: configuration for energy item
		:return: energy item
		:raises ValueError: if item could not be found
		"""
		if isinstance(data, HABApp.openhab.items.NumberItem):
			return data
		try:
			return HABApp.openhab.items.NumberItem.get_item(data)
		except HABApp.core.errors.ItemNotFoundException:
			raise ValueError(f"Could not find any item for given name '{data}'")


class MonthlyReportItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for monthly report."""
	energy_sum: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="item which holds the total energy consumption")


class MonthlyReportParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for monthly report."""
	known_energy_shares: list[EnergyShare] = pydantic.Field([], description="list of EnergyShare objects which hold the known energy shares. E.g. energy for lights or ventilation")
	persistence_group_name: str | None = pydantic.Field(None, description="OpenHAB group name which holds all items which are persisted. If the group name is given it will be checked if all energy items are in the group")
	config_mail: multi_notifier.connectors.connector_mail.MailConfig = pydantic.Field(..., description="config for sending mails")
	recipients: list[str] = pydantic.Field(..., description="list of recipients who get the mail")
	debug: bool = pydantic.Field(False, description="if debug mode is active")


class MonthlyReportConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for monthly report."""
	items: MonthlyReportItems = pydantic.Field(..., description="Items for monthly report")
	parameter: MonthlyReportParameter = pydantic.Field(..., description="Parameter for monthly report")



"""Rules to handle sun sensors."""
import logging

import HABApp

import habapp_rules.common.config.filter
import habapp_rules.common.filter
import habapp_rules.common.hysteresis
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.sun

LOGGER = logging.getLogger(__name__)


class _SensorBase(HABApp.Rule):
	"""Base class for sun sensors."""

	def __init__(self, config: habapp_rules.sensors.config.sun.BrightnessConfig | habapp_rules.sensors.config.sun.TemperatureDifferenceConfig, item_input: HABApp.openhab.items.NumberItem) -> None:
		"""Init of base class for sun sensors.

		:param config: config for sun sensor
		:param item_input: item for input value (brightness or temperature difference)
		"""
		self._config = config

		# init HABApp Rule
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.output.name)

		# init exponential filter
		name_input_exponential_filtered = f"H_{item_input.name.removeprefix('H_')}_filtered"
		habapp_rules.core.helper.create_additional_item(name_input_exponential_filtered, "Number", name_input_exponential_filtered.replace("_", " "), config.parameter.filtered_signal_groups)
		item_input_filtered = HABApp.openhab.items.NumberItem.get_item(name_input_exponential_filtered)

		exponential_filter_config = habapp_rules.common.config.filter.ExponentialFilterConfig(
			items=habapp_rules.common.config.filter.ExponentialFilterItems(
				raw=item_input,
				filtered=item_input_filtered),
			parameter=habapp_rules.common.config.filter.ExponentialFilterParameter(
				tau=config.parameter.filter_tau,
				instant_increase=config.parameter.filter_instant_increase,
				instant_decrease=config.parameter.filter_instant_decrease
			)
		)
		habapp_rules.common.filter.ExponentialFilter(exponential_filter_config)

		# attributes
		self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(config.threshold, config.parameter.hysteresis, False)

		# callbacks
		item_input_filtered.listen_event(self._cb_input_filtered, HABApp.openhab.events.ItemStateChangedEventFilter())
		if config.items.threshold is not None:
			config.items.threshold.listen_event(self._cb_threshold, HABApp.openhab.events.ItemStateChangedEventFilter())

	def _send_output(self, new_value: str) -> None:
		"""Send output if different.

		:param new_value: new value which should be sent
		"""
		if new_value != self._config.items.output.value:
			self._config.items.output.oh_send_command(new_value)
			self._instance_logger.debug(f"Set output '{self._config.items.output.name}' to {new_value}")

	def _cb_input_filtered(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the filtered input value changed.

		:param event: trigger event
		"""
		value = self._hysteresis_switch.get_output(event.value)
		self._send_output(value)

	def _cb_threshold(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the threshold changed.

		:param event: trigger event
		"""
		self._hysteresis_switch.set_threshold_on(event.value)


class SensorBrightness(_SensorBase):
	"""Rules class to set sun protection depending on brightness level.

	# Items:
	Number    brightness                    "Current brightness [%d lux]"       {channel="..."}
	Number    brightness_threshold          "Brightness threshold [%d lux]"
	Switch    sun_protection_brightness     "Sun protection brightness [%s]     {channel="..."}

	# Config:
	config = habapp_rules.sensors.config.sun.BrightnessConfig(
		items=habapp_rules.sensors.config.sun.BrightnessItems(
			brightness="brightness",
			output="sun_protection_brightness",
			brightness_threshold="brightness_threshold"
		)
	)

	# Rule init:
	habapp_rules.sensors.sun.SensorBrightness(config)
	"""

	def __init__(self, config: habapp_rules.sensors.config.sun.BrightnessConfig) -> None:
		"""Init of sun sensor which takes a brightness value

		:param config: config for the sun sensor which is using brightness
		"""
		_SensorBase.__init__(self, config, config.items.brightness)


class SensorTemperatureDifference(_SensorBase):
	"""Rules class to set sun protection depending on temperature difference. E.g. temperature in the sun / temperature in the shadow.

	# Items:
	Number    temperature_sun               "Temperature sun [%.1f °C]"         {channel="..."}
	Number    temperature_shadow            "Temperature shadow [%.1f °C]"      {channel="..."}
	Number    temperature_threshold         "Temperature threshold [%.1f °C]"
	Switch    sun_protection_temperature    "Sun protection temperature [%s]    {channel="..."}

	# Config:
	config = habapp_rules.sensors.config.sun.TemperatureDifferenceConfig(
		items=habapp_rules.sensors.config.sun.TemperatureDifferenceItems(
			temperatures=["temperature_sun", "temperature_shadow"],
			output="sun_protection_temperature",
			threshold="temperature_threshold"
		)
	)

	# Rule init:
	habapp_rules.sensors.sun.SensorTempDiff(config)
	"""

	def __init__(self, config: habapp_rules.sensors.config.sun.TemperatureDifferenceConfig) -> None:
		"""Init of sun sensor which takes a two or more temperature values (one in the sun and one in the shadow)

		:param config: config for the sun sensor which is using temperature items
		"""
		self._config = config
		name_temperature_diff = f"H_Temperature_diff_for_{config.items.output.name}"
		habapp_rules.core.helper.create_additional_item(name_temperature_diff, "Number", name_temperature_diff.replace("_", " "))
		self._item_temp_diff = HABApp.openhab.items.NumberItem.get_item(name_temperature_diff)

		_SensorBase.__init__(self, config, self._item_temp_diff)

		# callbacks
		for temperature_item in self._config.items.temperatures:
			temperature_item.listen_event(self._cb_temperature, HABApp.openhab.events.ItemStateChangedEventFilter())

		# calculate temperature difference
		self._cb_temperature(None)

	def _cb_temperature(self, _: HABApp.openhab.events.ItemStateChangedEvent | None):
		"""Callback, which is triggered if a temperature value changed."""
		filtered_items = [itm for itm in habapp_rules.core.helper.filter_updated_items(self._config.items.temperatures, self._config.parameter.ignore_old_values_time) if itm.value is not None]
		if len(filtered_items) < 2:
			return
		value_min = min(item.value for item in filtered_items)
		value_max = max(item.value for item in filtered_items)

		self._item_temp_diff.oh_send_command(value_max - value_min)


class SunPositionFilter(HABApp.Rule):
	"""Rules class to filter a switch state depending on the sun position. This can be used to only close the blinds of a window, if the sun hits the window

	# Items:
	Number    sun_azimuth           "Sun Azimuth [%.1f °]"              {channel="astro..."}
	Number    sun_elevation         "Sun Elevation [%.1f °]"            {channel="astro..."}
	Switch    sun_shining           "Sun is shining [%s]
	Switch    sun_hits_window       "Sun hits window [%s]

	# Config:
	config = habapp_rules.sensors.config.sun.SunPositionConfig(
		items=habapp_rules.sensors.config.sun.SunPositionItems(
			azimuth="sun_azimuth",
			elevation="sun_elevation",
			input="sun_shining",
			output="sun_hits_window"
		),
		parameter=habapp_rules.sensors.config.sun.SunPositionParameter(
			sun_position_window=habapp_rules.sensors.config.sun.SunPositionWindow(40, 120)
		)
	)

	# Rule init:

	habapp_rules.sensors.sun.SunPositionFilter(config)
	"""

	def __init__(self, config: habapp_rules.sensors.config.sun.SunPositionConfig) -> None:
		"""Init of sun position filter.

		:param config: config for the sun position filter
		"""
		self._config = config

		# init HABApp Rule
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.output.name)

		# callbacks
		config.items.azimuth.listen_event(self._update_output, HABApp.openhab.events.ItemStateChangedEventFilter())  # listen_event for elevation is not needed because elevation and azimuth is updated together
		config.items.input.listen_event(self._update_output, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._update_output(None)

	def _sun_in_window(self, azimuth: float, elevation: float) -> bool:
		"""Check if the sun is in the 'sun window' where it hits the target.

		:param azimuth: azimuth of the sun
		:param elevation: elevation of the sun
		:return: True if the sun hits the target, else False
		"""
		sun_in_window = False

		if any(window.azimuth_min <= azimuth <= window.azimuth_max and window.elevation_min <= elevation <= window.elevation_max for window in self._config.parameter.sun_position_windows):
			sun_in_window = True

		return sun_in_window

	def _update_output(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
		"""Callback, which is triggered if the sun position or input changed."""
		azimuth = self._config.items.azimuth.value
		elevation = self._config.items.elevation.value

		if azimuth is None or elevation is None:
			self._instance_logger.warning(f"Azimuth or elevation is None -> will set output to input. azimuth = {azimuth} | elevation = {elevation}")
			filter_output = self._config.items.input.value
		elif self._config.items.input.value in {"OFF", None}:
			filter_output = "OFF"
		else:
			filter_output = "ON" if self._sun_in_window(azimuth, elevation) else "OFF"

		if filter_output != self._config.items.output.value:
			self._config.items.output.oh_send_command(filter_output)



"""Config models for light rules."""
import collections.abc
import logging
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.exceptions
import habapp_rules.core.pydantic_base

LOGGER = logging.getLogger(__name__)


class LightItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for all light rules."""
	light: HABApp.openhab.items.DimmerItem | HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which controls the light")
	light_control: list[HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="control items to improve manual detection")
	light_groups: list[HABApp.openhab.items.DimmerItem] = pydantic.Field([], description="group items which can additionally set the light state. This can be used to improve the manual detection")
	manual: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item to switch to manual mode and disable the automatic functions")
	presence_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="presence state set via habapp_rules.presence.Presence")
	day: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="item which is ON at day and OFF at night")
	sleeping_state: HABApp.openhab.items.StringItem | None = pydantic.Field(None, description="sleeping state set via habapp_rules.system.sleep.Sleep")
	pre_sleep_prevent: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="item to prevent pre-sleep (Can be used for example to prevent the pre sleep light when guests are sleeping)")
	doors: list[HABApp.openhab.items.ContactItem] = pydantic.Field([], description="door items for switching on the light if the door is opening")
	motion: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="motion sensor to enable light if motion is detected")
	state: HABApp.openhab.items.StringItem = pydantic.Field(..., description="item to store the current state of the state machine")


class BrightnessTimeout(pydantic.BaseModel):
	"""Define brightness and timeout for light states."""
	brightness: int | bool = pydantic.Field(..., description="brightness which should be set. If bool ON will be sent for True and OFF for False")
	timeout: float = pydantic.Field(..., description="Timeout / max time in seconds until switch off")

	def __init__(self, brightness: int | bool, timeout: float):
		"""Initialize BrightnessTimeout without kwargs

		:param brightness: brightness value
		:param timeout: timeout value
		"""
		super().__init__(brightness=brightness, timeout=timeout)

	@pydantic.model_validator(mode="after")
	def validata_model(self) -> typing.Self:
		"""Validate brightness and timeout

		:raises AssertionError: if brightness and timeout are not valid
		:return: self
		"""
		if self.brightness is False or self.brightness == 0:
			# Default if the light should be switched off e.g. for leaving / sleeping
			if not self.timeout:
				self.timeout = 0.5

		if not self.timeout:
			raise AssertionError(f"Brightness and timeout are not valid: brightness = {self.brightness} | timeout = {self.timeout}")
		return self


class FunctionConfig(pydantic.BaseModel):
	"""Define brightness and timeout values for one function."""
	day: BrightnessTimeout | None = pydantic.Field(..., description="config for day. If None the light will not be switched on during the day")
	night: BrightnessTimeout | None = pydantic.Field(..., description="config for night. If None the light will not be switched on during the night")
	sleeping: BrightnessTimeout | None = pydantic.Field(..., description="config for sleeping. If None the light will not be switched on during sleeping")


class LightParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for all light rules.

	For all parameter which have the following type: FunctionConfig | None -> If None this state will be disabled for the rule
	"""
	on: FunctionConfig = pydantic.Field(FunctionConfig(day=BrightnessTimeout(True, 14 * 3600), night=BrightnessTimeout(80, 10 * 3600), sleeping=BrightnessTimeout(20, 3 * 3600)),
	                                    description="values which are used if the light is switched on manually")
	pre_off: FunctionConfig | None = pydantic.Field(FunctionConfig(day=BrightnessTimeout(50, 10), night=BrightnessTimeout(40, 7), sleeping=BrightnessTimeout(10, 7)), description="values which are used if the light changes pre_off state")
	leaving: FunctionConfig | None = pydantic.Field(FunctionConfig(day=BrightnessTimeout(False, 0), night=BrightnessTimeout(False, 0), sleeping=BrightnessTimeout(False, 0)), description="values which are used if the light changes to leaving state")
	pre_sleep: FunctionConfig | None = pydantic.Field(FunctionConfig(day=BrightnessTimeout(False, 10), night=BrightnessTimeout(False, 10), sleeping=None), description="values which are used if the light changes to pre_sleep state")
	pre_sleep_prevent: collections.abc.Callable[[], bool] | HABApp.openhab.items.OpenhabItem | None = pydantic.Field(None, description="Enable pre sleep prevent -> disable pre sleep if True")
	motion: FunctionConfig | None = pydantic.Field(None, description="values which are used if the light changes to motion state")
	door: FunctionConfig | None = pydantic.Field(None, description="values which are used if the light is enabled via a door opening")
	off_at_door_closed_during_leaving: bool = pydantic.Field(False, description="this can be used to switch lights off, when door is closed in leaving state")
	hand_off_lock_time: int = pydantic.Field(20, description="time in seconds where door / motion switch on is disabled after a manual OFF")
	leaving_only_if_on: bool = pydantic.Field(False, description="switch to leaving only if light is on. If False leaving light is always activated")

	@pydantic.field_validator("on", mode="after")
	@classmethod
	def validate_on(cls, value: FunctionConfig) -> FunctionConfig:
		"""Validate config for on-state

		:param value: given value
		:return: validated value
		:raises AssertionError: if on is not valid
		"""
		if any(conf is None for conf in [value.day, value.night, value.sleeping]):
			raise AssertionError("For function 'on' all brightness / timeout values must be set.")
		return value

	@pydantic.field_validator("pre_sleep", mode="after")
	@classmethod
	def validate_pre_sleep(cls, value: FunctionConfig | None) -> FunctionConfig | None:
		"""Validate pre_sleep config

		:param value: value of pre sleep
		:raises AssertionError: if pre_sleep is not valid
		:return: validated value
		"""
		if value is None:
			return value

		if value.sleeping is not None:
			LOGGER.warning("It's not allowed to set brightness / timeout for pre_sleep.sleeping. Set it to None")
			value.sleeping = None

		return value


class LightConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for all light rules."""
	items: LightItems = pydantic.Field(..., description="items for all light rules")
	parameter: LightParameter = pydantic.Field(LightParameter(), description="parameter for all light rules")

	@pydantic.model_validator(mode="after")
	def validate_config(self) -> typing.Self:
		"""Validate config

		:raises AssertionError: if config is not valid
		:return: config if valid
		"""
		if self.items.motion is not None and self.parameter.motion is None:
			raise AssertionError("item motion is given, but not configured via parameter")

		if len(self.items.doors) and self.parameter.door is None:
			raise AssertionError("item door is given, but not configured via parameter")

		if self.items.sleeping_state is not None and self.parameter.pre_sleep is None:
			raise AssertionError("item sleeping_state is given, but not configured via parameter")

		if self.items.presence_state is not None and self.parameter.leaving is None:
			raise AssertionError("item presence_state is given, but not configured via parameter")

		if self.items.pre_sleep_prevent is not None and self.parameter.pre_sleep_prevent is not None:
			LOGGER.warning("item pre_sleep_prevent and parameter pre_sleep_prevent are given. The item will be prioritized and the parameter will be ignored!")

		return self



"""Config models for KNX / MQTT bridge."""
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class KnxMqttItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Configuration of items for KNX MQTT bridge."""
	mqtt_dimmer: HABApp.openhab.items.DimmerItem = pydantic.Field(..., description="")
	knx_switch_ctr: HABApp.openhab.items.SwitchItem | None = pydantic.Field(None, description="")
	knx_dimmer_ctr: HABApp.openhab.items.DimmerItem | None = pydantic.Field(None, description="")

	@pydantic.model_validator(mode="after")
	def validate_knx_items(self) -> typing.Self:
		"""Validate KNX items

		:return: validated model
		:raises ValueError: if knx_switch_ctr and knx_dimmer_ctr are not set
		"""
		if self.knx_switch_ctr is None and self.knx_dimmer_ctr is None:
			raise ValueError("knx_switch_ctr or knx_dimmer_ctr must be set")
		return self


class KnxMqttParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Configuration of parameters for KNX MQTT bridge."""
	increase_value: int = pydantic.Field(60, description="")
	decrease_value: int = pydantic.Field(30, description="")


class KnxMqttConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Configuration of KNX MQTT bridge."""
	items: KnxMqttItems = pydantic.Field(..., description="Items for KNX MQTT bridge")
	parameter: KnxMqttParameter = pydantic.Field(KnxMqttParameter(), description="Parameters for KNX MQTT bridge")



"""Config models for filter rules."""
import typing

import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class ExponentialFilterItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for exponential filter."""
	raw: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Item for raw value")
	filtered: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Item for filtered value")


class ExponentialFilterParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for exponential filter."""
	tau: int = pydantic.Field(..., description="filter time constant in seconds. E.g. step from 0 to 1 | tau = 5 seconds -> after 5 seconds the value will be 0,67")
	instant_increase: bool = pydantic.Field(False, description="if set to True, increase of input values will not be filtered")
	instant_decrease: bool = pydantic.Field(False, description="if set to True, decrease of input values will not be filtered")

	@pydantic.model_validator(mode="after")
	def validate_instant_parameters(self) -> typing.Self:
		"""Validate instant_increase and instant_decrease.

		:return: validated model
		:raises ValueError: if both parameters are set
		"""
		if self.instant_decrease and self.instant_increase:
			raise ValueError("instant_increase and instant_decrease can not be set to True at the same time!")
		return self


class ExponentialFilterConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for exponential filter."""
	items: ExponentialFilterItems = pydantic.Field(..., description="Items for exponential filter")
	parameter: ExponentialFilterParameter = pydantic.Field(..., description="Parameter for exponential filter")



"""Rules for managing motion sensors."""
import logging

import HABApp

import habapp_rules.common.hysteresis
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.sensors.config.motion
import habapp_rules.system.sleep

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member, too-many-instance-attributes
class Motion(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Class for filtering motion sensors.

	# MQTT-things:
	Thing topic Motion "Motion Sensor"{
        Channels:
        Type switch : motion        "Motion"        [stateTopic="zigbee2mqtt/Motion/occupancy", on="true", off="false"]
        Type number : brightness    "Brightness"    [stateTopic="zigbee2mqtt/Motion/illuminance_lux"]
    }

    # Items:
    Switch    Motion_raw                "Motion raw"                <motion>        {channel="mqtt:topic:broker:Motion:motion"}
	Switch    Motion_filtered           "Motion filtered"           <motion>
	Number    Motion_Brightness         "Brightness"                                {channel="mqtt:topic:broker:Motion:brightness"}
	String    I999_00_Sleeping_state    "Sleeping state"            <state>

	# Config
	config = habapp_rules.sensors.config.motion.MotionConfig(
		items=habapp_rules.sensors.config.motion.MotionItems(
			motion_raw="Motion_raw",
			motion_filtered="Motion_filtered",
			brightness="Motion_Brightness",
			sleep_state="I999_00_Sleeping_state"
		),
		parameter=habapp_rules.sensors.config.motion.MotionParameter(
			brightness_threshold=100,
		),
	)

	# Rule init:
	habapp_rules.sensors.motion.Motion(config)
	"""
	states = [
		{"name": "Locked"},
		{"name": "SleepLocked"},
		{"name": "PostSleepLocked", "timeout": 99, "on_timeout": "timeout_post_sleep_locked"},
		{"name": "Unlocked", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "Wait"},
			{"name": "Motion"},
			{"name": "MotionExtended", "timeout": 99, "on_timeout": "timeout_motion_extended"},
			{"name": "TooBright"},
		]}
	]

	trans = [
		# lock
		{"trigger": "lock_on", "source": ["Unlocked", "SleepLocked", "PostSleepLocked"], "dest": "Locked"},
		{"trigger": "lock_off", "source": "Locked", "dest": "Unlocked", "unless": "_sleep_active"},
		{"trigger": "lock_off", "source": "Locked", "dest": "SleepLocked", "conditions": "_sleep_active"},

		# sleep
		{"trigger": "sleep_started", "source": ["Unlocked", "PostSleepLocked"], "dest": "SleepLocked"},
		{"trigger": "sleep_end", "source": "SleepLocked", "dest": "Unlocked", "unless": "_post_sleep_lock_configured"},
		{"trigger": "sleep_end", "source": "SleepLocked", "dest": "PostSleepLocked", "conditions": "_post_sleep_lock_configured"},
		{"trigger": "timeout_post_sleep_locked", "source": "PostSleepLocked", "dest": "Unlocked", "unless": "_raw_motion_active"},
		{"trigger": "motion_off", "source": "PostSleepLocked", "dest": "PostSleepLocked"},
		{"trigger": "motion_on", "source": "PostSleepLocked", "dest": "PostSleepLocked"},

		# motion
		{"trigger": "motion_on", "source": "Unlocked_Wait", "dest": "Unlocked_Motion"},
		{"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_MotionExtended", "conditions": "_motion_extended_configured"},
		{"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_Wait", "unless": "_motion_extended_configured"},
		{"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Wait", "unless": "_brightness_over_threshold"},
		{"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_TooBright", "conditions": "_brightness_over_threshold"},
		{"trigger": "motion_on", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Motion"},

		# brightness
		{"trigger": "brightness_over_threshold", "source": "Unlocked_Wait", "dest": "Unlocked_TooBright"},
		{"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Wait", "unless": "_raw_motion_active"},
		{"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Motion", "conditions": "_raw_motion_active"}
	]

	# pylint: disable=too-many-arguments
	def __init__(self, config: habapp_rules.sensors.config.motion.MotionConfig) -> None:
		"""Init of motion filter.

		:param config: config for the motion filter
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not valid
		"""
		self._config = config
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.motion_raw.name)

		self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(threshold := self._get_brightness_threshold(), threshold * 0.1 if threshold else 5) if self._config.items.brightness is not None else None

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self.state_machine.get_state("PostSleepLocked").timeout = self._config.parameter.post_sleep_lock_time
		self.state_machine.get_state("Unlocked_MotionExtended").timeout = self._config.parameter.extended_motion_time

		# register callbacks
		self._config.items.motion_raw.listen_event(self._cb_motion_raw, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.brightness is not None:
			self._config.items.brightness.listen_event(self._cb_brightness, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.brightness_threshold is not None:
			self._config.items.brightness_threshold.listen_event(self._cb_threshold_change, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.lock is not None:
			self._config.items.lock.listen_event(self._cb_lock, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._config.items.sleep_state is not None:
			self._config.items.sleep_state.listen_event(self._cb_sleep, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = "initial") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._config.items.lock is not None and self._config.items.lock.is_on():
			return "Locked"
		if self._config.items.sleep_state is not None and self._config.items.sleep_state.value == habapp_rules.system.SleepState.SLEEPING.value:
			return "SleepLocked"
		if self._config.items.brightness is not None and self._brightness_over_threshold():
			return "Unlocked_TooBright"
		if self._config.items.motion_raw.is_on():
			return "Unlocked_Motion"
		return "Unlocked_Wait"

	def _update_openhab_state(self):
		"""Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self.__send_filtered_motion()

			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")
			self._previous_state = self.state

	def __send_filtered_motion(self) -> None:
		"""Send filtered motion state to OpenHAB item."""
		target_state = "ON" if self.state in {"Unlocked_Motion", "Unlocked_MotionExtended"} else "OFF"
		if target_state != self._config.items.motion_filtered.value:
			self._config.items.motion_filtered.oh_send_command(target_state)

	def _raw_motion_active(self) -> bool:
		"""Check if raw motion is active

		:return: True if active, else False
		"""
		return bool(self._config.items.motion_raw)

	def _brightness_over_threshold(self) -> bool:
		"""Check if brightness is over threshold

		:return: True if active, else False
		"""
		return self._hysteresis_switch.get_output(self._config.items.brightness.value)

	def _motion_extended_configured(self) -> bool:
		"""Check if extended motion is configured

		:return: True if active, else False
		"""
		return self._config.parameter.extended_motion_time > 0

	def _post_sleep_lock_configured(self) -> bool:
		"""Check if post sleep lock is configured

		:return: True if active, else False
		"""
		return self._config.parameter.post_sleep_lock_time > 0

	def _sleep_active(self) -> bool:
		"""Check if sleeping is active

		:return: True if sleeping is active, else False
		"""
		return self._config.items.sleep_state.value == habapp_rules.system.SleepState.SLEEPING.value

	def _get_brightness_threshold(self) -> float:
		"""Get the current brightness threshold value.
		
		:return: brightness threshold
		:raises habapp_rules.core.exceptions.HabAppRulesException: if brightness value not given by item or value
		"""
		if self._config.parameter.brightness_threshold:
			return self._config.parameter.brightness_threshold
		if self._config.items.brightness_threshold is not None:
			return value if (value := self._config.items.brightness_threshold.value) else float("inf")
		raise habapp_rules.core.exceptions.HabAppRulesException(f"Can not get brightness threshold. Brightness value or item is not given. value: {self._config.parameter.brightness_threshold} | item: {self._config.items.brightness_threshold}")

	# pylint: disable=invalid-name
	def on_enter_Unlocked_Init(self):
		"""Callback, which is called on enter of Unlocked_Init state"""
		if self._config.items.brightness is not None and self._brightness_over_threshold():
			self.to_Unlocked_TooBright()
		elif self._config.items.motion_raw.is_on():
			self.to_Unlocked_Motion()
		else:
			self.to_Unlocked_Wait()

	def _check_brightness(self, value: float | None = None) -> None:
		"""Check if brightness is higher than the threshold and trigger the class methods.

		:param value: Value to check. None if last value should be used
		"""
		if self._hysteresis_switch.get_output(value):
			self.brightness_over_threshold()
		else:
			self.brightness_below_threshold()

	def _cb_threshold_change(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the brightness threshold state changed.

		:param event: trigger event
		"""
		self._hysteresis_switch.set_threshold_on(event.value)
		self._check_brightness()

	def _cb_motion_raw(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the raw motion state changed.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.motion_on()
		else:
			self.motion_off()

	def _cb_brightness(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the brightness state changed.

		:param event: trigger event
		"""
		self._check_brightness(event.value)

	def _cb_lock(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the lock state changed.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.lock_on()
		else:
			self.lock_off()

	def _cb_sleep(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the sleep state changed.

		:param event: trigger event
		"""
		if event.value == habapp_rules.system.SleepState.SLEEPING.value:
			self.sleep_started()
		if event.value == habapp_rules.system.SleepState.AWAKE.value:
			self.sleep_end()



"""Implementations of logical functions."""
import abc
import logging

import HABApp

import habapp_rules.common.config.logic
import habapp_rules.core.helper
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class _BinaryLogicBase(HABApp.Rule):
	"""Base class for binary logical functions."""

	def __init__(self, config: habapp_rules.common.config.logic.BinaryLogicConfig) -> None:
		"""Init a logical function.

		:param config: Config for logical function
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		HABApp.Rule.__init__(self)
		self._config = config
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{self._config.items.output.name}")

		if isinstance(self._config.items.output, HABApp.openhab.items.SwitchItem):
			# item type is Switch
			self._positive_state = "ON"
			self._negative_state = "OFF"
		else:
			# item type is Contact (validated by type of config)
			self._positive_state = "CLOSED"
			self._negative_state = "OPEN"

		for item in self._config.items.inputs:
			item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateUpdatedEventFilter())

		self._cb_input_event(None)
		input_names = [item.name for item in self._config.items.inputs]
		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{self._config.items.output.name}' | Input items = '{input_names}'")

	@abc.abstractmethod
	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""

	def _set_output_state(self, output_state: str) -> None:
		"""Set state to the output element

		:param output_state: state which will be set
		"""
		if isinstance(self._config.items.output, HABApp.openhab.items.ContactItem):
			self._config.items.output.oh_post_update(output_state)
		else:
			habapp_rules.core.helper.send_if_different(self._config.items.output, output_state)


class And(_BinaryLogicBase):
	"""Logical AND function.

	# Config:
	config = habapp_rules.common.config.logic.BinaryLogicConfig(
		items=habapp_rules.common.config.logic.BinaryLogicItems(
			inputs=["Item_1", "Item_2"],
			output="Item_result",
		)
	)

	# Rule init:
	habapp_rules.common.logic.And(config)
	"""

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		output_state = self._positive_state if all(item.value == self._positive_state for item in self._config.items.inputs) else self._negative_state
		self._set_output_state(output_state)


class Or(_BinaryLogicBase):
	"""Logical OR function.

	# Config:
	config = habapp_rules.common.config.logic.BinaryLogicConfig(
		items=habapp_rules.common.config.logic.BinaryLogicItems(
			inputs=["Item_1", "Item_2"],
			output="Item_result",
		)
	)

	# Rule init:
	habapp_rules.common.logic.Or(config)
	"""

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		output_state = self._positive_state if any(item.value == self._positive_state for item in self._config.items.inputs) else self._negative_state
		self._set_output_state(output_state)


class _NumericLogicBase(HABApp.Rule):
	"""Base class for numeric logical functions."""

	def __init__(self, config: habapp_rules.common.config.logic.NumericLogicConfig) -> None:
		"""Init a logical function.

		:param config: Config for logical function
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		HABApp.Rule.__init__(self)
		self._config = config
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{self._config.items.output.name}")

		for item in self._config.items.inputs:
			item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._cb_input_event(None)
		input_names = [item.name for item in self._config.items.inputs]
		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{self._config.items.output.name}' | Input items = '{input_names}'")

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		filtered_items = habapp_rules.core.helper.filter_updated_items(self._config.items.inputs, self._config.parameter.ignore_old_values_time)
		value = self._apply_numeric_logic([item.value for item in filtered_items if item is not None])

		if value is None:
			return

		self._set_output_state(value)

	@abc.abstractmethod
	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: value which fulfills the filter type
		"""

	def _set_output_state(self, output_state: float) -> None:
		"""Set state to the output element

		:param output_state: state which will be set
		"""
		habapp_rules.core.helper.send_if_different(self._config.items.output, output_state)


class Min(_NumericLogicBase):
	"""Logical Min function with filter for old / not updated items.

	# Config:
	config = habapp_rules.common.config.logic.NumericLogicConfig(
		items=habapp_rules.common.config.logic.NumericLogicItems(
			inputs=["Item_1", "Item_2"],
			output="Item_result",
		),
		parameter=habapp_rules.common.config.logic.NumericLogicParameter(
			ignore_old_values_time=600
		),
	)

	# Rule init:
	habapp_rules.common.logic.Min(config)
	"""

	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: min value of the given values
		"""
		return HABApp.util.functions.min(input_values)


class Max(_NumericLogicBase):
	"""Logical Max function with filter for old / not updated items.

	# Config:
	config = habapp_rules.common.config.logic.NumericLogicConfig(
		items=habapp_rules.common.config.logic.NumericLogicItems(
			inputs=["Item_1", "Item_2"],
			output="Item_result",
		),
		parameter=habapp_rules.common.config.logic.NumericLogicParameter(
			ignore_old_values_time=600
		),
	)

	# Rule init:
	habapp_rules.common.logic.Max(config)
	"""

	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: max value of the given values
		"""
		return HABApp.util.functions.max(input_values)


class Sum(_NumericLogicBase):
	"""Logical Sum function with filter for old / not updated items.

	# Config:
	config = habapp_rules.common.config.logic.NumericLogicConfig(
		items=habapp_rules.common.config.logic.NumericLogicItems(
			inputs=["Item_1", "Item_2"],
			output="Item_result",
		),
		parameter=habapp_rules.common.config.logic.NumericLogicParameter(
			ignore_old_values_time=600
		),
	)

	# Rule init:
	habapp_rules.common.logic.Sum(config)
	"""

	def __init__(self, config: habapp_rules.common.config.logic.NumericLogicConfig) -> None:
		"""Init a logical function.

		:param config: config for logical sum rule
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		if isinstance(config.items.output, HABApp.openhab.items.DimmerItem):
			raise TypeError(f"Dimmer items can not be used for Sum function! Given output_name: {config.items.output}")

		_NumericLogicBase.__init__(self, config)

	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: min value of the given values
		"""
		return sum(val for val in input_values if val is not None)


class InvertValue(HABApp.Rule):
	"""Rule to update another item if the value of an item changed.

	# Config:
	config = habapp_rules.common.config.logic.InvertValueConfig(
		items=habapp_rules.common.config.logic.InvertValueItems(
			input="Item_1",
			output="Item_2",
		)
	)

	# Rule init:
	habapp_rules.common.logic.InvertValue(config)
	"""

	def __init__(self, config: habapp_rules.common.config.logic.InvertValueConfig) -> None:
		"""Init rule.

		:param config: Config for invert value rule
		"""
		HABApp.Rule.__init__(self)
		self._config = config
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{self._config.items.output.name}")

		self._config.items.input.listen_event(self._cb_input_value, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._cb_input_value(HABApp.openhab.events.ItemStateChangedEvent(self._config.items.input.name, self._config.items.input.value, None))
		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{self._config.items.output.name}' | Input item = '{self._config.items.input.name}'")

	def _cb_input_value(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Set output, when input value changed

		:param event: event, which triggered this callback
		"""
		if event.value is None:
			return

		output_value = -1 * event.value

		if self._config.parameter.only_negative and output_value > 0:
			output_value = 0
		elif self._config.parameter.only_positive and output_value < 0:
			output_value = 0

		self._config.items.output.oh_send_command(output_value)



"""Module to create donut charts."""
import collections.abc
import pathlib

import matplotlib.pyplot


def _auto_percent_format(values: list[float]) -> collections.abc.Callable:
	"""Get labels for representing the absolute value.

	:param values: list of all values
	:return: function which returns the formatted string if called
	"""

	def my_format(pct: float) -> str:
		"""get formatted value.

		:param pct: percent value
		:return: formatted value
		"""
		total = sum(values)
		return f"{(pct * total / 100.0):.1f} kWh"

	return my_format


def create_chart(labels: list[str], values: list[float], chart_path: pathlib.Path) -> None:
	"""Create the donut chart.

	:param labels: labels for the donut chart
	:param values: values of the donut chart
	:param chart_path: target path for the chart
	"""
	_, ax = matplotlib.pyplot.subplots()
	_, texts, _ = ax.pie(values, labels=labels, autopct=_auto_percent_format(values), pctdistance=0.7, textprops={"fontsize": 10})
	for text in texts:
		text.set_backgroundcolor("white")

	matplotlib.pyplot.savefig(str(chart_path), bbox_inches="tight", transparent=True)



"""Rule to detect presence or absence."""
import logging
import threading

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.items
import HABApp.util

import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system.config.presence
from habapp_rules.system import PresenceState

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class Presence(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Rules class to manage presence of a home.

	Hint: If you have some kind of guest-mode, use a guest-available switch as a phone to enable a persistent presence, also if all phones are not at home

	Example OpenHAB configuration:
	# KNX-things:
	Thing device T00_99_OpenHab_Presence "KNX OpenHAB Presence"{
        Type switch-control        : presence       "Presence"      [ ga="0/2/11+0/2/10"]
        Type switch-control        : leaving        "Leaving"       [ ga="0/2/21+0/2/20"]
    }

    # Items:
    Switch    I01_00_Presence    "Presence [%s]"    <presence>    (G00_00_rrd4j)    ["Status", "Presence"]    {channel="knx:device:bridge:T00_99_OpenHab_Presence:presence"}
	Switch    I01_00_Leaving     "Leaving [%s]"     <leaving>                                                 {channel="knx:device:bridge:T00_99_OpenHab_Presence:leaving"}

	# Config:
	config = habapp_rules.system.config.presence.PresenceConfig(
		items=habapp_rules.system.config.presence.PresenceItems(
			presence="I01_00_Presence",
			leaving="I01_00_Leaving"
		)
	)

	# Rule init:
	habapp_rules.system.presence.Presence(config)
	"""

	states = [
		{"name": "presence"},
		{"name": "leaving", "timeout": 5 * 60, "on_timeout": "absence_detected"},  # leaving takes 5 minutes
		{"name": "absence", "timeout": 1.5 * 24 * 3600, "on_timeout": "long_absence_detected"},  # switch to long absence after 1.5 days
		{"name": "long_absence"}
	]

	trans = [
		{"trigger": "presence_detected", "source": ["absence", "long_absence"], "dest": "presence"},
		{"trigger": "leaving_detected", "source": ["presence", "absence", "long_absence"], "dest": "leaving"},
		{"trigger": "abort_leaving", "source": "leaving", "dest": "presence"},
		{"trigger": "absence_detected", "source": ["presence", "leaving"], "dest": "absence"},
		{"trigger": "long_absence_detected", "source": "absence", "dest": "long_absence"},
	]

	def __init__(self, config: habapp_rules.system.config.presence.PresenceConfig) -> None:
		"""Init of Presence object.

		:param config: config for presence detection
		"""
		self._config = config
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, config.items.state)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.presence.name)

		# init state machine
		self.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		# add callbacks
		self._config.items.leaving.listen_event(self._cb_leaving, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._config.items.presence.listen_event(self._cb_presence, HABApp.openhab.events.ItemStateChangedEventFilter())
		for door_item in self._config.items.outdoor_doors:
			door_item.listen_event(self._cb_outside_door, HABApp.core.events.ValueChangeEventFilter())
		for phone_item in self._config.items.phones:
			phone_item.listen_event(self._cb_phone, HABApp.core.events.ValueChangeEventFilter())

		self.__phone_absence_timer: threading.Timer | None = None
		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = PresenceState.PRESENCE.value) -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: return correct state if it could be detected, if not return default value
		"""
		phone_items = [phone for phone in self._config.items.phones if phone.value is not None]  # phones with valid state (not None)
		if phone_items:
			if any((item.value == "ON" for item in phone_items)):
				return PresenceState.PRESENCE.value

			if self._config.items.presence.value == "ON":
				return PresenceState.LEAVING.value
			return PresenceState.LONG_ABSENCE.value if self._item_state.value == PresenceState.LONG_ABSENCE.value else PresenceState.ABSENCE.value

		if self._config.items.leaving.value == "ON":
			return PresenceState.LEAVING.value

		if self._config.items.presence.value == "ON":
			return PresenceState.PRESENCE.value

		if self._config.items.presence.value == "OFF":
			return PresenceState.LONG_ABSENCE.value if self._item_state.value == PresenceState.LONG_ABSENCE.value else PresenceState.ABSENCE.value

		return default_value

	def _update_openhab_state(self) -> None:
		"""Extend _update_openhab state of base class to also update other OpenHAB items."""
		super()._update_openhab_state()
		self._instance_logger.info(f"Presence state changed to {self.state}")

		# update presence item
		target_value = "ON" if self.state in {PresenceState.PRESENCE.value, PresenceState.LEAVING.value} else "OFF"
		habapp_rules.core.helper.send_if_different(self._config.items.presence, target_value)
		habapp_rules.core.helper.send_if_different(self._config.items.leaving, "ON" if self.state == PresenceState.LEAVING.value else "OFF")

	def _cb_outside_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if any outside door changed state.

		:param event: state change event of door item
		"""
		if event.value == "OPEN" and self.state not in {PresenceState.PRESENCE.value, PresenceState.LEAVING.value}:
			self._instance_logger.debug(f"Presence detected by door ({event.name})")
			self.presence_detected()

	def _cb_leaving(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if leaving item changed state.

		:param event: Item state change event of leaving item
		"""
		if event.value == "ON" and self.state in {PresenceState.PRESENCE.value, PresenceState.ABSENCE.value, PresenceState.LONG_ABSENCE.value}:
			self._instance_logger.debug("Start leaving through leaving switch")
			self.leaving_detected()
		if event.value == "OFF" and self.state == PresenceState.LEAVING.value:
			self._instance_logger.debug("Abort leaving through leaving switch")
			self.abort_leaving()

	def _cb_presence(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if presence item changed state.

		:param event: Item state change event of presence item
		"""
		if event.value == "ON" and self.state in {PresenceState.ABSENCE.value, PresenceState.LONG_ABSENCE.value}:
			self._instance_logger.debug("Presence was set manually by presence switch")
			self.presence_detected()
		elif event.value == "OFF" and self.state in {PresenceState.PRESENCE.value, PresenceState.LEAVING.value}:
			self._instance_logger.debug("Absence was set manually by presence switch")
			self.absence_detected()

	def _cb_phone(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is called if a phone state changed.

		:param event: Item state change event of phone item
		"""
		active_phones = len([phone for phone in self._config.items.phones if phone.value == "ON"])
		if active_phones == 1 and event.value == "ON":
			# first phone switched to ON
			if self.__phone_absence_timer:
				self.__phone_absence_timer.cancel()
				self.__phone_absence_timer = None

			if self.state == PresenceState.LEAVING.value:
				self._instance_logger.debug("Leaving was aborted through first phone which came online")
				self.abort_leaving()

			if self.state in {PresenceState.ABSENCE.value, PresenceState.LONG_ABSENCE.value}:
				self._instance_logger.debug("Presence was set through first phone joined network")
				self.presence_detected()

		elif active_phones == 0 and event.value == "OFF":
			# last phone switched to OFF
			self.__phone_absence_timer = threading.Timer(20 * 60, self.__set_leaving_through_phone)
			self.__phone_absence_timer.start()

	def __set_leaving_through_phone(self) -> None:
		"""Set leaving detected if timeout expired."""
		if self.state == PresenceState.PRESENCE.value:
			self._instance_logger.debug("Leaving was set, because last phone left some time ago.")
			self.leaving_detected()
		self.__phone_absence_timer = None

	def on_rule_removed(self) -> None:
		habapp_rules.core.state_machine_rule.StateMachineRule.on_rule_removed(self)

		# stop phone absence timer
		if self.__phone_absence_timer:
			self.__phone_absence_timer.cancel()
			self.__phone_absence_timer = None



"""Config models for sun rules."""
import logging
import typing

import HABApp
import pydantic

import habapp_rules.core.pydantic_base

LOGGER = logging.getLogger(__name__)


class _ItemsBase(habapp_rules.core.pydantic_base.ItemBase):
	"""Base class for items for sun sensor."""
	output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="output item")
	threshold: HABApp.openhab.items.NumberItem | None = pydantic.Field(None, description="threshold item")


class BrightnessItems(_ItemsBase):
	"""Items for sun sensor which uses brightness items as input."""
	brightness: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="brightness item")


class TemperatureDifferenceItems(_ItemsBase):
	"""Items for sun sensor which uses temperature items as input."""
	temperatures: list[HABApp.openhab.items.NumberItem] = pydantic.Field(..., description="temperature items")

	@pydantic.model_validator(mode="after")
	def validate_temperature_items(self) -> typing.Self:
		"""Validate that at least two temperature items are given.

		:return: validated model
		:raises ValueError: if less than two temperature items are given
		"""
		if len(self.temperatures) < 2:
			raise ValueError("At least two temperature items are required!")
		return self


class BrightnessParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for sun sensor which uses brightness items as input."""
	threshold: float | None = pydantic.Field(None, description="threshold value")
	hysteresis: float = pydantic.Field(0.0, description="hysteresis value")
	filter_tau: int = pydantic.Field(30 * 60, description="filter constant for the exponential filter. Default is set to 30 minutes")
	filter_instant_increase: bool = pydantic.Field(True, description="if set to True, increase of input values will not be filtered")
	filter_instant_decrease: bool = pydantic.Field(False, description="if set to True, decrease of input values will not be filtered")
	filtered_signal_groups: list[str] = pydantic.Field([], description="group names where the filtered signal will be added")


class TemperatureDifferenceParameter(BrightnessParameter):
	"""Parameter for sun sensor which uses temperature items as input."""
	ignore_old_values_time: int | None = pydantic.Field(None, description="ignores values which are older than the given time in seconds. If None, all values will be taken")


class _ConfigBase(habapp_rules.core.pydantic_base.ConfigBase):
	"""Base config model for sun sensor."""
	items: BrightnessItems | TemperatureDifferenceItems = pydantic.Field(..., description="items for sun sensor")
	parameter: BrightnessParameter | TemperatureDifferenceParameter = pydantic.Field(..., description="parameter for sun sensor")

	@pydantic.model_validator(mode="after")
	def validate_threshold(self) -> typing.Self:
		"""Validate threshold.

		:return: validated model
		:raises ValueError: if threshold and parameter are not set
		"""
		if (self.items.threshold is None) == (self.parameter.threshold is None):
			raise ValueError("The threshold must be set ether with the parameter or with the item, both are not allowed")
		return self

	@property
	def threshold(self) -> float:
		"""Get threshold."""
		if self.parameter.threshold:
			return self.parameter.threshold

		if self.items.threshold.value is None:
			LOGGER.warning("Threshold item has no value set. Setting threshold to infinity")
			return float("inf")

		return self.items.threshold.value


class BrightnessConfig(_ConfigBase):
	"""Config model for sun sensor which uses brightness as input."""
	items: BrightnessItems = pydantic.Field(..., description="items for sun sensor which uses brightness as input")
	parameter: BrightnessParameter = pydantic.Field(BrightnessParameter(), description="parameter for sun sensor which uses brightness as input")


class TemperatureDifferenceConfig(_ConfigBase):
	"""Config model for sun sensor which uses temperature items as input."""
	items: TemperatureDifferenceItems = pydantic.Field(..., description="items for sun sensor which uses temperature items as input")
	parameter: TemperatureDifferenceParameter = pydantic.Field(TemperatureDifferenceParameter(), description="parameter for sun sensor which uses temperature items as input")

############################ SunPositionFilter ###############################
class SunPositionWindow(pydantic.BaseModel):
	"""Class for defining min / max values for azimuth and elevation."""
	azimuth_min: float = pydantic.Field(..., description="Starting value for azimuth", ge=0.0, le=360.0)
	azimuth_max: float = pydantic.Field(..., description="End value for azimuth", ge=0.0, le=360.0)
	elevation_min: float = pydantic.Field(0.0, description="Starting value for elevation", ge=-90.0, le=90.0)
	elevation_max: float = pydantic.Field(90.0, description="End value for elevation", ge=-90.0, le=90.0)

	def __init__(self, azimuth_min: float, azimuth_max: float, elevation_min: float = 0.0, elevation_max: float = 90.0) -> None:
		"""Init of class for defining min / max values for azimuth and elevation.

		:param azimuth_min: minimum azimuth value
		:param azimuth_max: maximum azimuth value
		:param elevation_min: minimum elevation value
		:param elevation_max: maximum elevation value
		"""
		super().__init__(azimuth_min=azimuth_min, azimuth_max=azimuth_max, elevation_min=elevation_min, elevation_max=elevation_max)

	@pydantic.model_validator(mode="after")
	def validate_model(self) -> typing.Self:
		"""Validate values.

		:return: validated model
		"""
		if self.azimuth_min > self.azimuth_max:
			LOGGER.warning(f"azimuth_min should be smaller than azimuth_max -> min / max will be swapped. Given values: azimuth_min = {self.azimuth_min} | azimuth_max = {self.azimuth_max}")
			min_orig = self.azimuth_min
			max_orig = self.azimuth_max
			self.azimuth_min = max_orig
			self.azimuth_max = min_orig

		if self.elevation_min > self.elevation_max:
			LOGGER.warning(f"elevation_min should be smaller than elevation_max -> min / max will be swapped. Given values: elevation_min = {self.elevation_min} | elevation_max = {self.elevation_max}")
			min_orig = self.elevation_min
			max_orig = self.elevation_max
			self.elevation_min = max_orig
			self.elevation_max = min_orig
		return self


class SunPositionItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for sun position filter."""
	azimuth: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="sun azimuth item")
	elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="sun elevation item")
	input: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="input item (sun protection required)")
	output: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="output item (sun protection required and sun in the configured azimuth / elevation window)")


class SunPositionParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for sun position filter."""
	sun_position_window: SunPositionWindow | list[SunPositionWindow] = pydantic.Field(..., description="sun position window, where the sun hits the target")

	@property
	def sun_position_windows(self) -> list[SunPositionWindow]:
		"""Get sun position windows."""
		return self.sun_position_window if isinstance(self.sun_position_window, list) else [self.sun_position_window]


class SunPositionConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config model for sun position filter."""
	items: SunPositionItems = pydantic.Field(..., description="items for sun position filter")
	parameter: SunPositionParameter = pydantic.Field(..., description="parameter for sun position filter")



"""Config models for astro rules."""
import HABApp.openhab.items
import pydantic

import habapp_rules.core.pydantic_base


class _NightDayItems(habapp_rules.core.pydantic_base.ItemBase):
	"""Items for night day."""
	elevation: HABApp.openhab.items.NumberItem = pydantic.Field(..., description="Elevation of the sun")


class SetDayItems(_NightDayItems):
	"""Items for setting day."""
	day: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item which should be set to ON after dawn and OFF after dusk")


class SetNightItems(_NightDayItems):
	"""Items for setting night."""
	night: HABApp.openhab.items.SwitchItem = pydantic.Field(..., description="Item which should be set to ON after dusk and OFF after dawn")


class SetDayParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for setting day."""
	elevation_threshold: float = pydantic.Field(0.0, description="Threshold value for elevation. If the sun elevation is greater than the threshold, the day item will be set to ON")


class SetNightParameter(habapp_rules.core.pydantic_base.ParameterBase):
	"""Parameter for setting night."""
	elevation_threshold: float = pydantic.Field(-8.0, description="Threshold value for elevation. If the sun elevation is greater than the threshold, the night item will be set to ON")


class SetDayConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for setting day."""
	items: SetDayItems = pydantic.Field(..., description="Items for setting day")
	parameter: SetDayParameter = pydantic.Field(SetDayParameter(), description="Parameter for setting day")


class SetNightConfig(habapp_rules.core.pydantic_base.ConfigBase):
	"""Config for setting night."""
	items: SetNightItems = pydantic.Field(..., description="Items for setting night")
	parameter: SetNightParameter = pydantic.Field(SetNightParameter(), description="Parameter for setting night")
